/*
 *
 * Date:        2009/11/30
 * Info:        mod_dnsblacklist Lighttpd module
 * Contact:     mailto: <luca [at] lucaercoli.it> http://www.lucaercoli.it
 * Version:     1.1
 * Author:      Luca Ercoli
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 */


#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "base.h"
#include "log.h"
#include "buffer.h"
#include "plugin.h"
#include "inet_ntop_cache.h"
#include <setjmp.h>
#include <signal.h>

#include <errno.h>
#include <resolv.h>
#include <arpa/nameser.h>

extern int h_errno;

typedef union {
	HEADER qb1;
	u_char qb2[PACKETSZ];
} querybuf;


typedef struct {
	buffer *dnsblacklist_method;
	buffer *dnsblacklist_host;
	buffer *dnsblacklist_message;
} plugin_config;

typedef struct {
	PLUGIN_DATA;

	plugin_config **config_storage;

	plugin_config conf;
} plugin_data;



INIT_FUNC(mod_dnsblacklist_init) {
	plugin_data *p;

	p = calloc(1, sizeof(*p));

	return p;
}

FREE_FUNC(mod_dnsblacklist_free) {
	plugin_data *p = p_d;

	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;

	if (p->config_storage) {
		size_t i;
		for (i = 0; i < srv->config_context->used; i++) {
			plugin_config *s = p->config_storage[i];

			if (!s) continue;
			buffer_free(s->dnsblacklist_method);
			buffer_free(s->dnsblacklist_host);
			buffer_free(s->dnsblacklist_message);

			free(s);
		}
		free(p->config_storage);
	}

	free(p);

	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_dnsblacklist_set_defaults) {
	plugin_data *p = p_d;
	size_t i = 0;

	config_values_t cv[] = {
		{ "dnsblacklist.method",             NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },
		{ "dnsblacklist.host",               NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },
		{ "dnsblacklist.message",            NULL, T_CONFIG_STRING, T_CONFIG_SCOPE_SERVER },
		{ NULL,                        NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
	};

	if (!p) return HANDLER_ERROR;

	p->config_storage = calloc(1, srv->config_context->used * sizeof(specific_config *));

	for (i = 0; i < srv->config_context->used; i++) {
		plugin_config *s;

		s = calloc(1, sizeof(plugin_config));
		s->dnsblacklist_method = buffer_init();
		s->dnsblacklist_host =  buffer_init();
		s->dnsblacklist_message = buffer_init();

		cv[0].destination = s->dnsblacklist_method;
		cv[1].destination = s->dnsblacklist_host;
		cv[2].destination = s->dnsblacklist_message;

		p->config_storage[i] = s;

		if (0 != config_insert_values_global(srv, ((data_config *)srv->config_context->data[i])->value, cv)) {
			return HANDLER_ERROR;
		}


	}

	p->conf.dnsblacklist_method = p->config_storage[0]->dnsblacklist_method;
	p->conf.dnsblacklist_host = p->config_storage[0]->dnsblacklist_host;
	p->conf.dnsblacklist_message = p->config_storage[0]->dnsblacklist_message;


	if (buffer_is_empty(p->conf.dnsblacklist_method)) {
		buffer_copy_string(p->conf.dnsblacklist_method,"POST");
	}

	if (buffer_is_empty(p->conf.dnsblacklist_host)) {
		buffer_copy_string(p->conf.dnsblacklist_host,"sbl-xbl.spamhaus.org");
	}

	if (buffer_is_empty(p->conf.dnsblacklist_message)) {
		buffer_copy_string(p->conf.dnsblacklist_message,"Your IP address is blacklisted!");
	}

	return HANDLER_GO_ON;
}

#define PATCH(x) \
	p->conf.x = s->x;
static int mod_dnsblacklist_patch_connection(server *srv, connection *con, plugin_data *p) {
	size_t i;
	plugin_config *s = p->config_storage[0];

	PATCH(dnsblacklist_method);
	PATCH(dnsblacklist_host);
	PATCH(dnsblacklist_message);

	for (i = 1; i < srv->config_context->used; i++) {
		data_config *dc = (data_config *)srv->config_context->data[i];
		s = p->config_storage[i];

		if (!config_check_cond(srv, con, dc)) continue;

	}

	return 0;
}
#undef PATCH

URIHANDLER_FUNC(mod_dnsblacklist_uri_handler) {

	plugin_data *p = p_d;

	char lookup_this[512];
	int oct1,oct2,oct3,oct4;

	mod_dnsblacklist_patch_connection(srv, con, p);

	char *metodo;

	switch (con->request.http_method) {
		case HTTP_METHOD_GET: metodo = "GET";
				      break;
		case HTTP_METHOD_POST: metodo = "POST";
				       break;
		case HTTP_METHOD_HEAD:  metodo = "HEAD";
					break;
		case HTTP_METHOD_OPTIONS: metodo = "OPTIONS";
					  break;
		case HTTP_METHOD_PUT:	 metodo = "PUT";
					 break;
		case HTTP_METHOD_PROPFIND: metodo="PROPFIND";
					   break;
		default: metodo="ALTRO"; 
			 break;
	}


	if ( buffer_search_string_len(p->conf.dnsblacklist_method, metodo, strlen(metodo) )  != NULL ){

		switch (con->dst_addr.plain.sa_family) {
			case AF_INET:

				memset(lookup_this,'\0',512);

				sscanf(inet_ntop_cache_get_ip(srv, &(con->dst_addr)), "%d.%d.%d.%d",&oct1, &oct2, &oct3, &oct4);

				snprintf(lookup_this,512,"%d.%d.%d.%d.%s",oct4,oct3,oct2,oct1,p->conf.dnsblacklist_host->ptr);

				_res.retry=1;
				_res.retrans=1;
				res_init();

				querybuf answer;
				int rc, num_answers;
				rc = res_query(lookup_this, C_IN,T_A,answer.qb2, sizeof(querybuf));

				num_answers = ntohs(answer.qb1.ancount);

				if (num_answers == 1){

					log_error_write(srv, __FILE__, __LINE__, "ss", "mod_dnsblacklist: Deny incoming connections by",inet_ntop_cache_get_ip(srv, &(con->dst_addr)));
					buffer *output = chunkqueue_get_append_buffer(con->write_queue);
					buffer_copy_string_len(output,p->conf.dnsblacklist_message->ptr,strlen(p->conf.dnsblacklist_message->ptr));
					con->file_finished = 1;

					return HANDLER_FINISHED;
				}


				break;
			default:
				return HANDLER_GO_ON;
		};

	}



	return HANDLER_GO_ON;
}


int mod_dnsblacklist_plugin_init(plugin *p) {
	p->version     = LIGHTTPD_VERSION_ID;
	p->name        = buffer_init_string("dnsblacklist");

	p->init        = mod_dnsblacklist_init;
	p->set_defaults = mod_dnsblacklist_set_defaults;
	p->handle_uri_clean  = mod_dnsblacklist_uri_handler;
	p->cleanup     = mod_dnsblacklist_free;

	p->data        = NULL;

	return 0;
}
